import logging


log = logging.getLogger(__name__)


class Result:
    PASS = True
    FAIL = False
    success = PASS

    def start(self, message):
        self.success = self.PASS

    def end(self):
        pass

    def fail(self, message, path=None):
        raise NotImplementedError()   # pragma: no cover


class LoggedResult(Result):
    def start(self, interaction):
        super().start(interaction)
        log.info(f'Verifying {interaction}')

    def fail(self, message, path=None):
        self.success = self.FAIL
        log.warning(' ' + message)
        return not message


class PytestResult(Result):   # pragma: no cover
    def start(self, interaction):
        log.info(f'Verifying {interaction}')

    def fail(self, message, path=None):
        from _pytest.outcomes import Failed
        __tracebackhide__ = True
        self.success = self.FAIL
        log.warning(' ' + message)
        raise Failed(message) from None

    def configure_logging(self, verbosity):
        logging.getLogger().handlers = []
        logging.basicConfig(format='%(message)s')
        log.setLevel({0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbosity])



import pytest

from pactman.verifier.parse_header import Part, parse_header


def test_comma_whitespace_ignored():
    assert list(parse_header('spam, ham')) == list(parse_header('spam,ham'))


@pytest.mark.parametrize(['header', 'result'], [
    (('audio/*; q=0.2, audio/basic'), [Part(['audio/*'], []), Part(['audio/basic'], [('q', '0.2')])]),
    (('''text/plain; q=0.5, text/html,
         text/x-dvi; q="0.8", text/x-c'''), [Part(['text/plain'], []),
                                             Part(['text/html', 'text/x-dvi'], [('q', '0.5')]),
                                             Part(['text/x-c'], [('q', '0.8')])]),
    (('"xyz\\"zy",W/"r2d2,xxxx","spam"'), [Part(['"xyz\\"zy"', 'W/"r2d2,xxxx"', '"spam"'], [])]),
])
def test_accept_variants(header, result):
    assert list(parse_header(header)) == result



import logging
from urllib.parse import parse_qs, urljoin

import requests

from .matching_rule import fold_type, nice_type, rule_matchers_v2, rule_matchers_v3, RuleFailed
from .parse_header import parse_header
from .paths import format_path


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class ProviderStateError(Exception):
    pass


class Interaction:
    def __init__(self, pact, interaction, result):
        self.pact = pact
        self.result = result
        self.description = interaction['description']
        self.request = interaction['request']
        self.providerState = interaction.get('providerState')
        self.providerStates = interaction.get('providerStates')
        self.response = ResponseVerifier(pact, interaction['response'], self.result)

    def __repr__(self):
        return f"{self.pact.consumer}:{self.description}"

    def verify(self, service_url, setup_url):
        self.result.start(self)
        try:
            self.result.success = self.setup(setup_url)
            if not self.result.success:
                return False
            self.result.success = self.run_service(service_url)
        finally:
            self.result.end()

    def run_service(self, service_url):
        method = self.request['method']
        handler = getattr(self, f'service_{method}', None)
        if handler is None:
            return self.result.fail(f'Request method {method} not implemented in verifier')
        return handler(service_url)

    def service_GET(self, service_url):
        if 'query' in self.request:
            query = self.request['query']
            if isinstance(query, str):
                # version 2 spec used strings, version 3 uses objects
                query = parse_qs(query)
            r = requests.get(self._get_url(service_url), params=query, headers=self._request_headers)
        else:
            r = requests.get(self._get_url(service_url), headers=self._request_headers)
        return self.response.verify(r)

    def service_POST(self, service_url):
        if not self._content_type_json:
            return self.result.fail(f'POST content type {self._content_type} not implemented in verifier')
        r = requests.post(self._get_url(service_url), json=self._request_payload, headers=self._request_headers)
        return self.response.verify(r)

    def service_DELETE(self, service_url):
        r = requests.delete(self._get_url(service_url), headers=self._request_headers)
        return self.response.verify(r)

    def service_PUT(self, service_url):
        if not self._content_type_json:
            return self.result.fail(f'PUT content type {self._content_type} not implemented in verifier')
        r = requests.put(self._get_url(service_url), json=self._request_payload, headers=self._request_headers)
        return self.response.verify(r)

    def service_PATCH(self, service_url):
        if not self._content_type_json:
            return self.result.fail(f'PATCH content type {self._content_type} not implemented in verifier')
        r = requests.patch(self._get_url(service_url), json=self._request_payload, headers=self._request_headers)
        return self.response.verify(r)

    def setup(self, setup_url):
        if self.providerState is not None:
            return self.set_up_state(setup_url, 'state', self.providerState)
        elif self.providerStates is not None:
            return self.set_up_state(setup_url, 'states', self.providerStates)
        return True

    def set_up_state(self, setup_url, var, state):
        log.debug(f'Setting up provider state {state!r}')
        args = {
            'provider': self.pact.provider,
            'consumer': self.pact.consumer,
            var: state
        }
        try:
            r = requests.post(setup_url, json=args)
        except requests.exceptions.ConnectionError as e:
            try:
                # try to pull the actual error out of the nested (nested (nested, string-ified))) exception
                reason = str(e.args[0].reason).split(": ", 1)[1]
            except Exception:
                reason = str(e)
            return self.result.fail(f'Unable to configure provider state {state!r} at {setup_url}: {reason}')
        if r.status_code != 200:
            log.debug(f'HTTP {r.status_code} from provider state setup URL: {r.text}')
            return self.result.fail(f'Invalid provider state {state!r}')
        log.info(f'Using provider state {state!r}')
        return True

    def _get_url(self, service_url):
        return urljoin(service_url, self.request['path'])

    @property
    def _request_headers(self):
        return self.request.get('headers', {})

    @property
    def _request_payload(self):
        return self.request['body']

    @property
    def _content_type_json(self):
        return 'application/json' in self._content_type

    @property
    def _content_type(self):
        return self._request_headers.get('Content-Type', 'application/json')


class ResponseVerifier:
    interaction_name = 'Response'

    def __init__(self, pact, interaction, result):
        self.pact = pact
        self.status = interaction.get('status', MISSING)
        self.headers = interaction.get('headers', MISSING)
        self.body = interaction.get('body', MISSING)
        rules = interaction.get('matchingRules', {})
        if pact.version['major'] < 2:
            # there are no matchingRules in v1
            self.matching_rules = {}
        elif pact.version['major'] < 3:
            self.matching_rules = rule_matchers_v2(rules)
        else:
            self.matching_rules = rule_matchers_v3(rules)
        self.result = result

    def verify(self, response):
        log.debug(f'{self.__class__.__name__}.verify headers={self.headers is not MISSING} '
                  f'body={self.body is not MISSING} rules={self.matching_rules != {}}')
        if self.status is not MISSING and response.status_code != self.status:
            log.debug(f'.. response {response.text}')
            return self.result.fail(f'{self.interaction_name} status code {response.status_code} is not '
                                    f'expected {self.status}')
        if self.headers is not MISSING:
            for header in self.headers:
                expected = self.headers[header]
                for actual in response.headers:
                    if header.lower() != actual.lower():
                        continue
                    actual = response.headers[actual]
                    if not self.check_rules(actual, expected, ['header', header]):
                        log.info(f'{self.interaction_name} headers: {response.json()}')
                        return False
        if self.body is not MISSING:
            if not self.check_rules(response.json(), self.body, ['body']):
                log.info(f'{self.interaction_name} data: {response.json()}')
                return False
        return True

    def check_rules(self, data, spec, path):
        log.debug(f'check_rules {data!r} {spec!r} {path}')
        if self.matching_rules:
            # if we have matchingRules then just look at those things
            r = self.apply_rules(data, spec, path)
        else:
            # otherwise the actual must equal the expected (excepting dict elements in actual that are unexpected)
            if path[0] == 'header':
                r = self.compare_header(data, spec, path)
            else:
                r = self.compare(data, spec, ['body'])
        log.debug(f'check_rules DONE = {r!r}')
        return r

    def compare_header(self, data, spec, path):
        parsed_data = sorted(parse_header(data))
        parsed_spec = sorted(parse_header(spec))
        log.debug(f'compare_header {parsed_data} {parsed_spec}')
        if parsed_data != parsed_spec:
            # there's a mind-bogglingly specific caveat to header matching that says that if the headers don't
            # match and they're a Content-Type and an encoding is supplier but not expected then it's OK to pass
            data_has_charset = [part for part in parsed_data if part.has_param('charset')]
            spec_has_charset = [part for part in parsed_spec if part.has_param('charset')]
            if path[1].lower() == 'content-type' and data_has_charset and not spec_has_charset:
                return True
            return self.result.fail(f'{self.interaction_name} header {path[1]} value {data!r} does not match '
                                    f'expected {spec!r}')
        return True

    def compare(self, data, spec, path):
        log.debug(f'compare {data!r} {spec!r} {path}')
        if fold_type(spec) is list:
            return self.compare_list(data, path, spec)
        if fold_type(spec) is dict:
            return self.compare_dict(data, spec, path)
        if not data == spec:
            return self.result.fail(f'Element mismatch {data!r} != {spec!r}', path)
        return True

    def compare_list(self, data, path, spec):
        if fold_type(data) is not list:
            return self.result.fail(f'{self.interaction_name} element is not an array (is {nice_type(data)})', path)
        if len(data) != len(spec):
            return self.result.fail(f'{self.interaction_name} array is incorrect length (is {len(data)} elements)',
                                    path)
        for i, (data_elem, spec_elem) in enumerate(zip(data, spec)):
            p = path + [i]
            if not self.compare(data_elem, spec_elem, p):
                return self.result.fail(f'{self.interaction_name} element {i} ({nice_type(data_elem)}) '
                                        f'does not match spec ({nice_type(spec_elem)})', path)
        return True

    def compare_dict(self, data, spec, path):
        if fold_type(data) is not dict:
            return self.result.fail(f'{self.interaction_name} element is not an object (is {nice_type(data)})', path)
        for key in spec:
            if key not in data:
                return self.result.fail(f'Expected element {key!r} not in response', path)
            p = path + [key]
            if not self.compare(data[key], spec[key], p):
                return self.result.fail(f'{self.interaction_name} element {key} ({nice_type(data[key])}) '
                                        f'does not match spec ({nice_type(spec[key])})', path)
        return True

    def apply_rules(self, data, spec, path):
        log.debug(f'apply_rules data={data!r} spec={spec!r} path={format_path(path)}')
        rule = self.find_rule(path)
        if rule:
            try:
                log.debug(f'... applying {rule}')
                rule.apply(data, spec, path)
            except RuleFailed as e:
                log.debug(f'... failed: {e}')
                return self.result.fail(str(e), path)
            log.debug('... passed')

        # we passed the matchingRule but we also need to check the contents of arrays/dicts
        if fold_type(spec) is list:
            r = self.apply_rules_array(data, spec, path)
            log.debug(f'apply_rules {format_path(path)} DONE = {r!r}')
            return r
        elif fold_type(spec) is dict:
            r = self.apply_rules_dict(data, spec, path)
            log.debug(f'apply_rules {format_path(path)} DONE = {r!r}')
            return r

        return True

    def find_rule(self, path):
        rules = self.matching_rules.get(path[0])
        if not rules:
            return None
        if self.pact.version['major'] > 2:
            rules = self.matching_rules.get(path[0])
            # version 3 rules paths don't include the interaction section ("body", "headers", ...)
            path = path[1:]
        weights = sorted((rule.weight(path), i, rule) for i, rule in enumerate(rules))
        weight, i, rule = weights[-1]
        if weight:
            return rule

    def apply_rules_array(self, data, spec, path):
        log.debug(f'apply_rules_array {data!r} {spec!r} {path!r}')
        if fold_type(data) is not list:
            return self.result.fail(f'{self.interaction_name} element is not an array (is {nice_type(data)})', path)
        # Attempt to find a matchingRule for this path elements in the array - if we find one then we use the first
        # spec value (since they must all satisfy the matchingRule and there may only be one) otherwise
        # we are comparing value to value so pass through the actual value from the spec.
        rule = self.find_rule(path + [0])
        if rule is not None:
            log.debug(f'... got a rule')
        else:
            log.debug(f'... performing elementwise rule application')
        for i, data_elem in enumerate(data):
            # always apply matching rules using the first element of the spec array
            spec_elem = spec[0]
            p = path + [i]
            if not self.apply_rules(data_elem, spec_elem, p):
                log.debug(f'apply_rules {path!r} failing on item {i}')
                return False
        log.debug(f'validate_type {path!r} DONE = True')
        return True

    def apply_rules_dict(self, data, spec, path):
        log.debug(f'apply_rules_dict {data!r} {spec!r} {path!r}')
        if fold_type(data) is not dict:
            return self.result.fail(f'{self.interaction_name} element is not an object (is {nice_type(data)})', path)
        for k in spec:
            p = path + [k]
            if k not in data:
                # we always flag a failure if a given key is not in the response
                return self.result.fail(f'Expected key {k!r} not in response', path)
            if not self.apply_rules(data[k], spec[k], p):
                return False
        return True


class RequestVerifier(ResponseVerifier):
    interaction_name = 'Request'

    def __init__(self, pact, interaction, result):
        self.method = interaction.get('method', MISSING)
        self.path = interaction.get('path', MISSING)
        self.query = interaction.get('query', MISSING)
        super().__init__(pact, interaction, result)

    def verify(self, request):
        log.debug(f'{self.__class__.__name__}.verify method={self.method is not MISSING} '
                  f'path={self.path is not MISSING} query={self.query is not MISSING}')
        if self.method is not MISSING and request.method.lower() != self.method.lower():
            return self.result.fail(f'Request method {request.method!r} does not match expected {self.method!r}')
        if self.path is not MISSING:
            if self.pact.version['major'] > 1 and self.matching_rules.get('path'):
                return self.apply_rules(request.path, self.path, ['path'])
            if request.path != self.path:
                return self.result.fail(f'Request path {request.path!r} does not match expected {self.path!r}')
        if self.query is not MISSING:
            spec_query = self.query
            if self.pact.version['major'] < 3:
                spec_query = parse_qs(spec_query)
            request_query = request.query
            if isinstance(request_query, str):
                request_query = parse_qs(request_query)
            if self.pact.version['major'] > 1 and self.matching_rules.get('query'):
                return self.apply_rules(request_query, spec_query, ['query'])
            if request_query != spec_query:
                return self.result.fail(f'Request query {request_query!r} does not match expected {spec_query!r}')
        return super().verify(request)

    def compare_dict(self, data, spec, path):
        if not super().compare_dict(data, spec, path):
            return False
        # check for unexpected data in the request
        for k in data:
            if k not in spec:
                return self.result.fail(f'Unexpected data in request', path + [k])
        return True


MISSING = object()



from unittest import TestCase
from unittest.mock import Mock

from pactman.mock.consumer import Consumer
from pactman.mock.provider import Provider
from pactman.mock.pact import Pact


class ConsumerTestCase(TestCase):
    def setUp(self):
        self.mock_service = Mock(Pact)
        self.provider = Mock(Provider)
        self.consumer = Consumer('TestConsumer', service_cls=self.mock_service)

    def test_init(self):
        result = Consumer('TestConsumer')
        self.assertIsInstance(result, Consumer)
        self.assertEqual(result.name, 'TestConsumer')
        self.assertIs(result.service_cls, Pact)

    def test_has_pact_with(self):
        result = self.consumer.has_pact_with(self.provider)
        self.assertIs(result, self.mock_service.return_value)
        self.mock_service.assert_called_once_with(
            consumer=self.consumer, provider=self.provider,
            host_name='localhost', port=1234,
            log_dir=None, ssl=False, sslcert=None, sslkey=None,
            pact_dir=None, version='2.0.0', use_mocking_server=False)

    def test_has_pact_with_customer_all_options(self):
        result = self.consumer.has_pact_with(
            self.provider, host_name='example.com', port=1111,
            log_dir='/logs', ssl=True,  sslcert='/ssl.cert', sslkey='ssl.pem',
            pact_dir='/pacts', version='3.0.0', use_mocking_server=False)

        self.assertIs(result, self.mock_service.return_value)
        self.mock_service.assert_called_once_with(
            consumer=self.consumer, provider=self.provider,
            host_name='example.com', port=1111,
            log_dir='/logs', ssl=True, sslcert='/ssl.cert', sslkey='ssl.pem',
            pact_dir='/pacts', version='3.0.0', use_mocking_server=False)

    def test_has_pact_with_not_a_provider(self):
        with self.assertRaises(ValueError):
            self.consumer.has_pact_with(None)



from unittest import TestCase

from pactman.mock.response import Response


class ResponseTestCase(TestCase):
    def test_sparse(self):
        target = Response(200)
        result = target.json('2.0.0')
        self.assertEqual(result, {'status': 200})

    def test_all_options(self):
        target = Response(
            202, headers={'Content-Type': 'application/json'}, body='the body')

        result = target.json('2.0.0')
        self.assertEqual(result, {
            'status': 202,
            'body': 'the body',
            'headers': {'Content-Type': 'application/json'}})

    def test_falsey_body(self):
        target = Response(200, body=[])
        result = target.json('2.0.0')
        self.assertEqual(result, {'status': 200, 'body': []})



import json
import logging
import queue
import traceback
from http.server import HTTPServer, BaseHTTPRequestHandler
from multiprocessing import Process, Queue

from .pact_request_handler import PactRequestHandler

_providers = {}


log = logging.getLogger(__name__)


def getMockServer(config):
    if config.provider_name not in _providers:
        _providers[config.provider_name] = Server(config)
    return _providers[config.provider_name]


class Server:
    def __init__(self, config):
        self.config = config
        self.interactions = Queue()
        self.results = Queue()
        self.process = Process(target=run_server, args=(config, self.interactions, self.results))
        self.process.start()

    def setup(self, interactions):
        for interaction in interactions:
            self.interactions.put_nowait(interaction)

    def verify(self):
        while not self.results.empty():
            result = self.results.get()
            if result['status'] == 'error':
                raise MockServer.Error(result['reason'])
            if result['status'] == 'failed':
                raise AssertionError(result['reason'])

    def terminate(self):
        self.process.terminate()


def run_server(config, interactions, results):
    httpd = MockServer(config, interactions, results)
    httpd.serve_forever()


class MockServer(HTTPServer):
    def __init__(self, config, interactions, results):
        self.config = config
        self.incoming_interactions = interactions
        self.outgoing_results = results
        server_address = ('', config.port)
        super().__init__(server_address, MockHTTPRequestHandler)
        self.interactions = []
        self.log = logging.getLogger(__name__ + '.' + config.provider_name)
        self.log.addHandler(logging.FileHandler(f'{config.log_dir}/{config.provider_name}.log'))
        self.log.setLevel(logging.DEBUG)
        self.log.propagate = False

    class Error(Exception):
        pass


class MockHTTPRequestHandler(BaseHTTPRequestHandler, PactRequestHandler):
    def __init__(self, request, client_address, server):
        self.response_status_code = None
        self.response_headers = {}
        self.response_body = None
        PactRequestHandler.__init__(self, server.config)
        BaseHTTPRequestHandler.__init__(self, request, client_address, server)

    def error_result(self, message, content='', status='error', status_code=500):
        self.server.outgoing_results.put({'status': status, 'reason': message})
        self.response_status_code = status_code
        self.response_headers = {'Content-Type': 'text/plain; charset=utf-8'}
        self.response_body = (content or message).encode('utf8')

    def run_request(self, method):
        try:
            self.body = None
            for header in self.headers:
                if header.lower() == 'content-length':
                    self.body = self.rfile.read(int(self.headers[header]))
            self.validate_request(method)
        except AssertionError as e:
            self.error_result(str(e))
        except Exception as e:
            self.error_result(f'Internal Error: {e}', traceback.format_exc())
        self.send_response(self.response_status_code)
        for header in self.response_headers:
            self.send_header(header, self.response_headers[header])
        self.end_headers()
        if self.response_body:
            self.wfile.write(self.response_body)

    def get_interaction(self, path):
        try:
            interaction = self.server.incoming_interactions.get(False)
        except queue.Empty:
            raise AssertionError(f'Request at {path} received but no interaction registered') from None
        return interaction

    def handle_success(self, interaction):
        self.server.outgoing_results.put({'status': 'success'})

    def handle_failure(self, reason):
        self.error_result(reason, status='failed', status_code=418)

    def respond_for_interaction(self, interaction):
        self.response_status_code = interaction['response']['status']
        if 'headers' in interaction['response']:
            self.response_headers.update(interaction['response']['headers'])
        if 'body' in interaction['response']:
            is_json = False
            for h in self.response_headers:
                if h.lower() == 'content-type':
                    is_json = self.response_headers[h].startswith('application/json')
                    break
            else:
                is_json = True
                self.response_headers['Content-Type'] = 'application/json; charset=utf-8'
            if is_json:
                self.response_body = json.dumps(interaction['response']['body']).encode('utf8')
            else:
                self.response_body = interaction['response']['body']

    def do_DELETE(self):
        self.run_request('DELETE')

    def do_GET(self):
        self.run_request('GET')

    def do_HEAD(self):
        self.run_request('HEAD')

    def do_POST(self):
        self.run_request('POST')

    def do_PUT(self):
        self.run_request('PUT')

    def log_message(self, format, *args):
        self.server.log.info("MockServer %s\n" % format % args)



"""API for creating a contract and configuring the mock service."""
from __future__ import unicode_literals

import os

from pactman.mock.request import Request
from pactman.mock.response import Response
from .mock_server import getMockServer
from .mock_urlopen import MockURLOpenHandler
from .pact_request_handler import Config


USE_MOCKING_SERVER = os.environ.get('PACT_USE_MOCKING_SERVER', 'no') == 'yes'


class Pact(object):
    """
    Represents a contract between a consumer and provider.

    Provides Python context handlers to configure the Pact mock service to
    perform tests on a Python consumer. For example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact.given('the echo service is available')
    ...  .upon_receiving('a request is made to the echo service')
    ...  .with_request('get', '/echo', query={'text': 'Hello!'})
    ...  .will_respond_with(200, body='Hello!'))
    >>> with pact:
    ...   requests.get(pact.uri + '/echo?text=Hello!')

    The GET request is made to the mock service, which will verify that it
    was a GET to /echo with a query string with a key named `text` and its
    value is `Hello!`. If the request does not match an error is raised, if it
    does match the defined interaction, it will respond with the text `Hello!`.
    """

    HEADERS = {'X-Pact-Mock-Service': 'true'}

    def __init__(self, consumer, provider, host_name='localhost', port=1234,
                 log_dir=None, ssl=False, sslcert=None, sslkey=None,
                 pact_dir=None, version='2.0.0',
                 file_write_mode='overwrite', use_mocking_server=USE_MOCKING_SERVER):
        """
        Constructor for Pact.

        :param consumer: The consumer for this contract.
        :type consumer: pact.Consumer
        :param provider: The provider for this contract.
        :type provider: pact.Provider
        :param host_name: The host name where the mock service is running.
        :type host_name: str
        :param port: The port number where the mock service is running.
        :type port: int
        :param log_dir: The directory where logs should be written. Defaults to
            the current directory.
        :type log_dir: str
        :param ssl: Flag to control the use of a self-signed SSL cert to run
            the server over HTTPS , defaults to False.
        :type ssl: bool
        :param sslcert: Path to a custom self-signed SSL cert file, 'ssl'
            option must be set to True to use this option. Defaults to None.
        :type sslcert: str
        :param sslkey: Path to a custom key and self-signed SSL cert key file,
            'ssl' option must be set to True to use this option.
            Defaults to None.
        :type sslkey: str
        :param pact_dir: Directory where the resulting pact files will be
            written. Defaults to the current directory.
        :type pact_dir: str
        :param version: The Pact Specification version to use, defaults to
            '2.0.0'.
        :type version: str
        :param file_write_mode: `overwrite` or `merge`. Use `merge` when
            running multiple mock service instances in parallel for the same
            consumer/provider pair. Ensure the pact file is deleted before
            running tests when using this option so that interactions deleted
            from the code are not maintained in the file. Defaults to
            `overwrite`.
        :type file_write_mode: str
        :param use_mocking_server: If True the mocking will be done using a
            HTTP server rather than patching urllib3 connections.
        :type use_mocking_server: bool
        """
        self.scheme = 'https' if ssl else 'http'
        self.consumer = consumer
        self.file_write_mode = file_write_mode
        self.host_name = host_name
        self.log_dir = log_dir or os.getcwd()
        self.pact_dir = pact_dir or os.getcwd()
        self.port = port
        self.provider = provider
        # TODO implement SSL
        self.ssl = ssl
        self.sslcert = sslcert
        self.sslkey = sslkey
        self.version = version
        self.use_mocking_server = use_mocking_server
        self._interactions = []
        self._mock_handler = None

    @property
    def uri(self):
        return '{scheme}://{host_name}:{port}'.format(host_name=self.host_name, port=self.port, scheme=self.scheme)

    def given(self, provider_state):
        """
        Define the provider state for this pact.

        When the provider verifies this contract, they will use this field to
        setup pre-defined data that will satisfy the response expectations.

        :param provider_state: The short sentence that is unique to describe
            the provider state for this contract.
        :type provider_state: basestring
        :rtype: Pact
        """
        self._interactions.insert(0, {'provider_state': provider_state})
        return self

    def setup(self):
        self._mock_handler.setup(self._interactions)

    def start_mocking(self):
        # TODO hmm, the config is looking a lot like this Pact instance...
        config = Config(self.consumer.name, self.provider.name, self.log_dir, self.pact_dir, self.file_write_mode,
                        self.version)
        self.port = config.port
        if self.use_mocking_server:
            self._mock_handler = getMockServer(config)
        else:
            # ain't no port, we're monkey-patching (but the URLs we generate still need to look correct)
            self._mock_handler = MockURLOpenHandler(config)

    def stop_mocking(self):
        self._mock_handler.terminate()

    # legacy pact-python API support
    start_service = start_mocking
    stop_service = stop_mocking

    def upon_receiving(self, scenario):
        """
        Define the name of this contract.

        :param scenario: A unique name for this contract.
        :type scenario: basestring
        :rtype: Pact
        """
        self._interactions[0]['description'] = scenario
        return self

    def verify(self):
        """
        Have the mock service verify all interactions occurred.

        Calls the mock service to verify that all interactions occurred as
        expected, and has it write out the contracts to disk.

        :raises AssertionError: When not all interactions are found.
        """
        try:
            self._mock_handler.verify()
        finally:
            # clear the interactions once we've attempted to verify, allowing re-use of the server
            self._interactions[:] = []

    def with_request(self, method, path, body=None, headers=None, query=None):
        """
        Define the request the request that the client is expected to perform.

        :param method: The HTTP method.
        :type method: str
        :param path: The path portion of the URI the client will access.
        :type path: str, Matcher
        :param body: The request body, can be a string or an object that will
            serialize to JSON, like list or dict, defaults to None.
        :type body: list, dict or None
        :param headers: The headers the client is expected to include on with
            this request. Defaults to None.
        :type headers: dict or None
        :param query: The query options the client is expected to send. Can be
            a dict of keys and values, or a URL encoded string.
            Defaults to None.
        :type query: dict, basestring, or None
        :rtype: Pact
        """
        self._interactions[0]['request'] = Request(
            method, path, body=body, headers=headers, query=query).json(self.version)
        return self

    def will_respond_with(self, status, headers=None, body=None):
        """
        Define the response the server is expected to create.

        :param status: The HTTP status code.
        :type status: int
        :param headers: All required headers. Defaults to None.
        :type headers: dict or None
        :param body: The response body, or a collection of Matcher objects to
            allow for pattern matching. Defaults to None.
        :type body: Matcher, dict, list, basestring, or None
        :rtype: Pact
        """
        self._interactions[0]['response'] = Response(status,
                                                     headers=headers,
                                                     body=body).json(self.version)
        return self

    def __enter__(self):
        """
        Handler for entering a Python context.

        Sets up the mock service to expect the client requests.
        """
        self.setup()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Handler for exiting a Python context.

        Calls the mock service to verify that all interactions occurred as
        expected, and has it write out the contracts to disk.
        """
        if (exc_type, exc_val, exc_tb) != (None, None, None):
            return

        self.verify()



from unittest import TestCase

from pactman import Term
from pactman.mock.request import Request


class RequestTestCase(TestCase):
    def test_sparse(self):
        target = Request('GET', '/path')
        result = target.json('2.0.0')
        self.assertEqual(result, {
            'method': 'GET',
            'path': '/path'})

    def test_all_options(self):
        target = Request(
            'POST', '/path',
            body='the content',
            headers={'Accept': 'application/json'},
            query='term=test')

        result = target.json('2.0.0')
        self.assertEqual(result, {
            'method': 'POST',
            'path': '/path',
            'body': 'the content',
            'headers': {'Accept': 'application/json'},
            'query': 'term=test'})

    def test_falsey_body(self):
        target = Request('GET', '/path', body=[])
        result = target.json('2.0.0')
        self.assertEqual(result, {
            'method': 'GET',
            'path': '/path',
            'body': []})

    def test_matcher_in_path_gets_converted(self):
        target = Request('GET', Term(r'\/.+', '/test-path'))
        result = target.json('2.0.0')
        self.assertEqual(result, {
            'method': 'GET',
            'path': '/test-path',
            'matchingRules': {
                '$.path': {
                    'regex': r'\/.+'
                }
            }
        })



import os

from setuptools import find_packages, setup


here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, "pactman", "__version__.py")) as f:
    exec(f.read(), about)


def read(filename):
    with open(os.path.join(here, filename), 'rb') as f:
        return f.read().decode('utf-8')


setup(
    name='pactman',
    version=about['__version__'],
    description=('Tools for creating and verifying consumer driven contracts'
                 ' using the Pact framework.'),
    long_description=read('README.md'),
    long_description_content_type='text/markdown',  # This is important!
    author='ReeceTech',
    author_email='richard.jones@reece.com.au',
    url='https://github.com/reecetech/pact-python',
    entry_points='''
        [console_scripts]
        pact-verifier=pactman.verifier.command_line:main
    ''',
    install_requires=[
        'pytest',
        'requests',
        'semver',
        'colorama',
        'restnavigator'
        # 'click',
    ],
    packages=find_packages(),
    license='MIT, Copyright (c) 2018 ReeceTech',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Testing :: Mocking',
        'Topic :: Software Development :: Testing :: Acceptance',
    ]
)



import argparse
import logging

from colorama import Fore, Style, init

from .broker_pact import BrokerPact, BrokerPacts, log
from .result import Result


# TODO: add these options, which exist in the ruby command line?
'''
@click.option(
    'username', '--pact-broker-username',
    help='Username for Pact Broker basic authentication.')
@click.option(
    'password', '--pact-broker-password',
    envvar='PACT_BROKER_PASSWORD',
    help='Password for Pact Broker basic authentication. Can also be specified'
         ' via the environment variable PACT_BROKER_PASSWORD')
@click.option(
    'header', '--custom-provider-header',
    envvar='CUSTOM_PROVIDER_HEADER',
    help='Header to add to provider state set up and '
         'pact verification requests. '
         'eg \'Authorization: Basic cGFjdDpwYWN0\'. '
         'May be specified multiple times.')
@click.option(
    'timeout', '-t', '--timeout',
    default=30,
    help='The duration in seconds we should wait to confirm verification'
         ' process was successful. Defaults to 30.',
    type=int)
'''

parser = argparse.ArgumentParser(description='Verify pact contracts')

parser.add_argument('provider_name', metavar='PROVIDER_NAME',
                    help='the name of the provider being verified')

parser.add_argument('provider_url', metavar='PROVIDER_URL',
                    help='the URL of the provider service')

parser.add_argument('provider_setup_url', metavar='PROVIDER_SETUP_URL',
                    help='the URL to provider state setup')

parser.add_argument('-b', '--broker-url', default=None,
                    help='the URL of the pact broker; may also be provided in PACT_BROKER_URL environment variable')

parser.add_argument('-l', '--local-pact-file', default=None,
                    help='path to a local pact file')

parser.add_argument('-c', '--consumer', default=None,
                    help='the name of the consumer to test')

parser.add_argument('-r', '--publish-verification-results', default=False, action='store_true',
                    help='send verification results to the pact broker')

parser.add_argument('-a', '--provider-app-version', default=None,
                    help='provider application version, required for results publication (same as -p)')

parser.add_argument('-p', '--provider-version', default=None,
                    help='provider application version, required for results publication (same as -a)')

parser.add_argument('-v', '--verbose', default=False, action='store_true',
                    help='output more information about the verification')

parser.add_argument('-q', '--quiet', default=False, action='store_true',
                    help='output less information about the verification')


class CaptureResult(Result):
    def __init__(self):
        log.handlers = [self]
        log.setLevel(logging.DEBUG)
        self.messages = []
        self.level = logging.WARNING
        self.current_consumer = None

    def start(self, interaction):
        super().start(interaction)
        self.messages[:] = []
        if self.current_consumer != interaction.pact.consumer:
            print(f'{Style.BRIGHT}Consumer: {interaction.pact.consumer}')
            self.current_consumer = interaction.pact.consumer
        print(f'Request: "{interaction.description}" ... ', end='')

    def end(self):
        if self.success:
            print(Fore.GREEN + 'PASSED')
        else:
            print(Fore.RED + 'FAILED')
        if self.messages:
            print((Fore.RESET + '\n').join(self.messages))

    def fail(self, message, path=None):
        self.success = self.FAIL
        if path:
            message += ' at ' + self.format_path(path)
        log.error(message)
        return not message

    def handle(self, record):
        color = ''
        if record.levelno >= logging.ERROR:
            color = Fore.RED
        elif record.levelno >= logging.WARNING:
            color = Fore.YELLOW
        self.messages.append(' ' + color + record.msg)


def main():
    init(autoreset=True)
    result = CaptureResult()
    args = parser.parse_args()
    provider_version = args.provider_version or args.provider_app_version
    if args.publish_verification_results and not provider_version:
        print('Provider version is required to publish results to the broker')
        return False
    if args.quiet:
        result.level = logging.WARNING
    elif args.verbose:
        result.level = logging.DEBUG
    else:
        result.level = logging.INFO
    if args.local_pact_file:
        pacts = [BrokerPact.load_file(args.local_pact_file, result)]
    else:
        pacts = BrokerPacts(args.provider_name, args.broker_url, result).consumers()
    success = True
    for pact in pacts:
        if args.consumer and pact.consumer != args.consumer:
            continue
        for interaction in pact.interactions:
            interaction.verify(args.provider_url, args.provider_setup_url)
            success = interaction.result.success and success
        if args.publish_verification_results:
            pact.publish_result(provider_version)
        else:
            print()
    return int(not success)


if __name__ == '__main__':
    import sys
    sys.exit(main())
